from git_autograder.answers.rules import AnswerRule
from git_autograder.answers import GitAutograderAnswersRecord


class HasExactValueRule(AnswerRule):
    def __init__(self, value: str, is_case_sensitive: bool = False) -> None:
        super().__init__()
        self.value = value
        self.is_case_sensitive = is_case_sensitive

    def apply(self, answer: GitAutograderAnswersRecord) -> None:
        expected = self.value.lower() if self.is_case_sensitive else self.value
        given = answer.answer.lower() if self.is_case_sensitive else answer.answer
        if given != expected:
            raise Exception(f"Answer for {answer.question} is not right.")



from dataclasses import dataclass
from datetime import datetime
import os
from typing import Optional
from git_autograder import GitAutograderRepo


@dataclass
class GitAutograderRepoContext:
    repo: GitAutograderRepo
    is_local: bool
    exercise_name: str
    started_at: datetime
    repo_path: Optional[str | os.PathLike] = None



import functools
import os
from datetime import datetime
from typing import Callable

import pytz

from git_autograder import GitAutograderOutput, GitAutograderRepo, GitAutograderStatus
from git_autograder.exception import (
    GitAutograderInvalidStateException,
    GitAutograderWrongAnswerException,
)


def autograder() -> Callable[
    [Callable[..., GitAutograderOutput]], Callable[..., GitAutograderOutput]
]:
    """
    Decorator to denote that a function is an autograder function.

    Initializes the GitAutograderRepo and provides it as an argument to the function.

    Handles the Git Autograder exceptions thrown by the GitAutograderRepo or from wrong
    answers.

    All outputs are saved directly to disk.
    """

    def inner(
        func: Callable[..., GitAutograderOutput],
    ) -> Callable[..., GitAutograderOutput]:
        @functools.wraps(func)
        def wrapper(*args, **kwargs) -> GitAutograderOutput:
            output = None
            try:
                is_local = os.environ.get("is_local", "false") == "true"
                exercise_name = os.environ.get("exercise_name")
                if exercise_name is None:
                    raise GitAutograderInvalidStateException(
                        "Missing exercise_name in environment", None, None, is_local
                    )
                repo = GitAutograderRepo(is_local=is_local, exercise_name=exercise_name)
                output = func(repo, *args, **kwargs)
            except (
                GitAutograderInvalidStateException,
                GitAutograderWrongAnswerException,
            ) as e:
                output = GitAutograderOutput(
                    exercise_name=e.exercise_name,
                    started_at=e.started_at,
                    completed_at=datetime.now(tz=pytz.UTC),
                    is_local=e.is_local,
                    comments=[e.message] if isinstance(e.message, str) else e.message,
                    status=e.status,
                )
            except Exception as e:
                # Unexpected exception
                output = GitAutograderOutput(
                    exercise_name=None,
                    started_at=None,
                    completed_at=None,
                    is_local=None,
                    comments=[str(e)],
                    status=GitAutograderStatus.ERROR,
                )

            assert output is not None
            output.save()
            return output

        return wrapper

    return inner



from typing import List
from git_autograder.exception import GitAutograderInvalidStateException
from git_autograder.repo_context import GitAutograderRepoContext


class BranchHelper:
    def __init__(self, ctx: GitAutograderRepoContext) -> None:
        self.ctx = ctx

    def track_remote_branches(self, remotes: List[str], strict: bool = False) -> None:
        if self.ctx.is_local:
            return

        tracked = {"main"}
        for remote in self.ctx.repo.repo.remote("origin").refs:
            for r in remotes:
                if r not in tracked or f"origin/{r}" != remote.name:
                    continue
                tracked.add(r)
                self.ctx.repo.repo.git.checkout("-b", r, f"origin/{r}")
                break

        missed_remotes = list(set(remotes).difference(tracked))
        if len(missed_remotes) > 0 and strict:
            raise GitAutograderInvalidStateException(
                f"Missing branches {', '.join(missed_remotes)} in submission",
                self.ctx.exercise_name,
                self.ctx.started_at,
                self.ctx.is_local,
            )

    def has_branch(self, branch: str) -> bool:
        return branch in self.ctx.repo.repo.heads



from json import JSONEncoder
from typing import Any
from datetime import datetime


class Encoder(JSONEncoder):
    def default(self, o: Any) -> Any:
        if isinstance(o, datetime):
            return o.timestamp()
        return super().default(o)



from git_autograder.answers.answers_record import GitAutograderAnswersRecord
from git_autograder.answers.answers import GitAutograderAnswers

__all__ = ["GitAutograderAnswersRecord", "GitAutograderAnswers"]



from git_autograder.answers.rules.answer_rule import AnswerRule
from git_autograder.answers.rules.has_exact_value_rule import HasExactValueRule
from git_autograder.answers.rules.not_empty_rule import NotEmptyRule
from git_autograder.answers.rules.has_exact_list_rule import HasExactListRule
from git_autograder.answers.rules.contains_list_rule import ContainsListRule

__all__ = [
    "AnswerRule",
    "HasExactValueRule",
    "NotEmptyRule",
    "HasExactListRule",
    "ContainsListRule",
]



from typing import List
from git_autograder.answers.rules import AnswerRule
from git_autograder.answers import GitAutograderAnswersRecord


class HasExactListRule(AnswerRule):
    def __init__(
        self, values: List[str], ordered: bool = False, is_case_sensitive: bool = False
    ) -> None:
        self.values = values
        self.ordered = ordered
        self.is_case_sensitive = is_case_sensitive

    def apply(self, answer: GitAutograderAnswersRecord) -> None:
        expected = (
            [v.lower() for v in self.values] if self.is_case_sensitive else self.values
        )
        given = (
            [v.lower() for v in answer.answer_as_list()]
            if self.is_case_sensitive
            else answer.answer_as_list()
        )
        if self.ordered and expected != given:
            raise Exception(
                f"Answer for {answer.question} does not contain all of the right answers. Ensure that they follow the order specified."
            )
        elif set(expected).intersection(set(given)) != len(expected):
            raise Exception(
                f"Answer for {answer.question} does not contain all of the right answers."
            )



from datetime import datetime
from typing import List, Optional, Union

from git_autograder import GitAutograderStatus


class GitAutograderException(Exception):
    def __init__(
        self,
        message: Union[str, List[str]],
        exercise_name: Optional[str],
        started_at: Optional[datetime],
        is_local: Optional[bool],
        status: GitAutograderStatus,
    ) -> None:
        super().__init__(message)

        self.message = message
        self.exercise_name = exercise_name
        self.started_at = started_at
        self.is_local = is_local
        self.status = status


class GitAutograderInvalidStateException(GitAutograderException):
    def __init__(
        self,
        message: str,
        exercise_name: Optional[str],
        started_at: Optional[datetime],
        is_local: Optional[bool],
    ) -> None:
        super().__init__(
            message,
            exercise_name,
            started_at,
            is_local,
            GitAutograderStatus.ERROR,
        )


class GitAutograderWrongAnswerException(GitAutograderException):
    def __init__(
        self,
        comments: List[str],
        exercise_name: Optional[str],
        started_at: datetime,
        is_local: bool,
    ) -> None:
        super().__init__(
            comments,
            exercise_name,
            started_at,
            is_local,
            GitAutograderStatus.UNSUCCESSFUL,
        )



from typing import List

from git_autograder.answers import GitAutograderAnswers
from git_autograder.answers.rules import AnswerRule
from git_autograder.exception import (
    GitAutograderInvalidStateException,
    GitAutograderWrongAnswerException,
)
from git_autograder.repo_context import GitAutograderRepoContext


class AnswersHelper:
    def __init__(
        self, ctx: GitAutograderRepoContext, answers: GitAutograderAnswers
    ) -> None:
        self.ctx = ctx
        self.answers = answers

    def validate_question(
        self, question: str, rules: List[AnswerRule]
    ) -> "AnswersHelper":
        answer = self.answers.get_by_question(question)
        if answer is None:
            raise GitAutograderInvalidStateException(
                f"Missing question {question} in answers file.",
                exercise_name=self.ctx.exercise_name,
                is_local=self.ctx.is_local,
                started_at=self.ctx.started_at,
            )

        for rule in rules:
            try:
                rule.apply(answer)
            except Exception as e:
                raise GitAutograderWrongAnswerException(
                    [str(e)],
                    exercise_name=self.ctx.exercise_name,
                    is_local=self.ctx.is_local,
                    started_at=self.ctx.started_at,
                )

        return self



from git_autograder.autograder import autograder
from git_autograder.test_utils import setup_autograder, set_env
from git_autograder.repo import GitAutograderRepo
from git_autograder.status import GitAutograderStatus
from git_autograder.output import GitAutograderOutput

__all__ = [
    "autograder",
    "setup_autograder",
    "set_env",
    "GitAutograderRepo",
    "GitAutograderStatus",
    "GitAutograderOutput",
]
